from leapp.workflows.api.v3.testapi import TestAPI
from leapp.workflows.api.v2.newapi import NewAPI
from leapp.workflows.api.v1.depcheck import DepCheckAPI1, DepCheckAPI2, DepCheckAPI3, DepCheckAPI4

